from random import sample
from string import ascii_lowercase

from django.contrib.auth.models import User
from django.db import models

MAX_FEEDBACK_PER_SUBMISSION = 3
SLUG_LENGTH = 16


def random_slug():
    return ''.join(sample(ascii_lowercase, SLUG_LENGTH))


class SubmissionType(models.Model):

    # Fields
    name = models.CharField(max_length=50)
    full_score = models.PositiveIntegerField(default=0)
    task_description = models.TextField()
    possible_solution = models.TextField()
    correction_guideline = models.TextField()

    class Meta:
        verbose_name = "SubmissionType"
        verbose_name_plural = "SubmissionType Set"

    def __str__(self):
        return self.name


class Student(models.Model):

    # Fields
    matrikel_no = models.PositiveIntegerField(unique=True, default=0)
    user = models.OneToOneField(
        User, on_delete=models.CASCADE,
        limit_choices_to={'groups__name': 'Students'},
    )

    class Meta:
        verbose_name = "Student"
        verbose_name_plural = "Student Set"

    def __str__(self):
        return "{} ({})".format(self.user.username, self.matrikel_no)


class Submission(models.Model):

    # Fields
    slug = models.SlugField(editable=False, unique=True, default=random_slug)
    seen = models.BooleanField(default=False)
    type = models.ForeignKey(
        SubmissionType,
        related_name='submissions'
    )
    status = models.OneToOneField(
        'SubmissionStatus',
        null=True, editable=False,
        related_name='submission',
        on_delete=models.CASCADE
    )
    text = models.TextField()
    pre_corrections = models.TextField(blank=True)
    final_feedback = models.OneToOneField('Feedback', null=True, blank=True)
    student = models.ForeignKey(Student)

    class Meta:
        verbose_name = "Submission"
        verbose_name_plural = "Submission Set"

    def save(self, *args, **kwargs):
        self.status = SubmissionStatus.objects.create()
        super(Submission, self).save(*args, **kwargs)

    def __str__(self):
        return "Submission of type '{}' from Student '{}'".format(
            self.type,
            self.student
        )

    def feedback_counter(self):
        return self.feedback_list.count()

    def feedback_limit_reached(self):
        return self.feedback_counter() == MAX_FEEDBACK_PER_SUBMISSION

    @classmethod
    def assign_tutor(cls, tutor):
        """Assigns a tutor to a submission

        A submission is not assigned to the specified tutor in the case
            1. the tutor already has a feedback in progress
            2. there is no more feedback to give

        Arguments:
            tutor {User} -- the tutor that should be assigned
        """
        # Get a submission from the submission set

        unfinished = Feedback.tutor_unfinished_feedback(tutor)
        if unfinished:
            return

        ready = cls.objects.filter(status__status=SubmissionStatus.READY)
        ready = ready.exclude(feedback_list__of_tutor=tutor)
        if not ready:
            return

        submission = ready[0]

        feedback = Feedback()
        feedback.of_tutor = tutor
        feedback.of_submission = submission
        feedback.save()

        # set the status to in progress
        if submission.feedback_counter() == MAX_FEEDBACK_PER_SUBMISSION:
            submission.status.status = SubmissionStatus.LIMIT_REACHED
        submission.status.save()


class SubmissionStatus(models.Model):

    READY, LIMIT_REACHED = range(2)
    STATUS = (
        (READY, 'READY'),
        (LIMIT_REACHED, 'LIMIT_REACHED'),
    )

    # Fields
    status = models.PositiveIntegerField(
        choices=STATUS,
        default=READY,
    )

    class Meta:
        verbose_name = "Submission Status"
        verbose_name_plural = "SubmissionStatus Set"

    def __str__(self):
        return "Status is {} - Count is {}".format(
            self.status,
            self.submission.feedback_counter()
        )


class Feedback(models.Model):
    # Fields
    text = models.TextField()
    slug = models.SlugField(editable=False, unique=True, default=random_slug)
    final = models.BooleanField(default=False)
    empty = models.BooleanField(default=True)
    score = models.PositiveIntegerField(default=0)
    of_submission = models.ForeignKey(
        Submission,
        related_name='feedback_list'
    )
    of_tutor = models.ForeignKey(
        User,
        related_name='corrected_submissions',
        limit_choices_to={'groups__name': 'Tutors'}
    )
    of_reviewer = models.ForeignKey(
        User,
        related_name='reviewed_submissions',
        limit_choices_to={'groups__name': 'Reviewers'},
        blank=True, null=True
    )

    class Meta:
        verbose_name = "Feedback"
        verbose_name_plural = "Feedback Set"

    def __str__(self):
        return 'Feedback for {}'.format(self.of_submission)

    def is_full_score(self):
        return self.of_submission.type.full_score == self.score

    def get_full_score(self):
        return self.of_submission.type.full_score

    @classmethod
    def tutor_unfinished_feedback(cls, user):
        """Gets only the feedback that is assigned and not final. A tutor
        should have only one feedback assigned that is not final

        Arguments:
            user {User} -- the tutor who formed the request

        Returns:
            Feedback -- the feedback or none if no feedback was assigned
        """
        tutor_feedback = cls.objects.filter(
            of_tutor=user, empty=True,
        )
        return tutor_feedback[0] if tutor_feedback else None

    def tutor_assigned_feedback(cls, user):
        """Gets all feedback that is assigned to the tutor inculuding
        all status cases.

        Returns:
            [list] -- a QuerySet of tasks that have been assigned to this tutor
        """
        tutor_feedback = cls.objects.filter(of_tutor=user)
        assert len(tutor_feedback.objects.filter(empty=True)) <= 1
        return tutor_feedback

    def finalize_feedback(self, user):
        """ Used to mark feedback as final (reviewed)

        This makes it uneditable by the tutor

        Arguments:
            user {[type]} -- [description]
        """
        self.final = True
        self.of_reviewer = user
        self.of_submission.final_feedback = self
        self.save()

    def unfinalize_feedback(self):
        """ Used to mark feedback as final (reviewed)

        This makes it uneditable by the tutor

        Arguments:
            user {[type]} -- [description]
        """
        self.final = False
        self.of_reviewer = None
        self.of_submission.final_feedback = None
        self.save()

    def unassign_tutor(self):
        self.of_submission.status.status = SubmissionStatus.READY
