from django.contrib.auth import authenticate, login, logout
from django.contrib.auth.decorators import login_required, user_passes_test
from django.http import Http404, HttpResponse, HttpResponseRedirect
from django.shortcuts import render
from django.urls import reverse
from django.utils.decorators import method_decorator
from django.views.generic import DetailView
from django.views.generic.edit import UpdateView

from core.forms import FeedbackForm
from core.models import Feedback, Submission, SubmissionType


# Create your views here.

def in_groups(user, group_list):
    return bool(user.groups.filter(name__in=group_list)) or user.is_superuser


def group_required(*group_names):
    """Requires user membership in at least one of the groups passed in."""
    def _in_groups(u):
        if u.is_authenticated():
            return in_groups(u, group_names)
    return user_passes_test(_in_groups)


def index(request):
    context = {
        'boldmessage': 'Delbert Grady says hey there world!',
        'submission_types': SubmissionType.objects.all(),
    }
    return render(request, 'core/index.html', context)


def user_login(request):

    if request.method == 'POST':
        username = request.POST['username']
        password = request.POST['password']

        user = authenticate(username=username, password=password)

        if user is not None:
            if user.is_active:
                login(request, user)
                return HttpResponseRedirect(reverse('start'))
            else:
                return HttpResponse("Your Grady account is disabled.")

        else:
            # Bad login details were provided. So we can't log the user in.
            print("Invalid login details: {0}, {1}".format(username, password))
            return HttpResponse("Invalid login details supplied.")
    else:
        return HttpResponseRedirect('/')


@login_required(login_url='/')
def user_logout(request):
    logout(request)
    return HttpResponseRedirect('/')


@login_required(login_url='/')
def user_home(request):
    if in_groups(request.user, ('Students', )):
        return student_view(request)
    elif in_groups(request.user, ('Tutors', )):
        return tutor_view(request)
    elif in_groups(request.user, ('Reviewers', )):
        return reviewer_view(request)
    else:
        return HttpResponseRedirect(reverse('index'))


@group_required('Tutors')
def tutor_view(request):
    context = {
        'username': request.user.username,
        'feedback_list': Feedback.objects.filter(of_tutor=request.user),
    }
    return render(request, 'core/tutor_startpage.html', context)


@group_required('Students')
def student_view(request):
    context = {
        'student': request.user,
        'submission_list': Submission.objects.filter(student__user=request.user)
    }
    return render(request, 'core/student_startpage.html', context)


@group_required('Reviewers')
def reviewer_view(request):
    context = {
        'reviewer': request.user,
        'submission_list': Submission.objects.all(),
        'feedback_list': Feedback.objects.all(),
    }
    return render(request, 'core/reviewer_startpage.html', context)


@group_required('Tutors', 'Reviewers')
def create_feedback(request):
    Submission.assign_tutor(request.user)
    feedback = Feedback.tutor_unfinished_feedback(request.user)
    if not feedback:
        return HttpResponseRedirect(reverse('end'))
    return HttpResponseRedirect('/feedback/edit/%s/' % feedback.slug)


@group_required('Tutors', 'Reviewers')
def delete_feedback(request, feedback_slug):
    """ Hook to ensure object is owned by request.user. """
    instance = Feedback.objects.get(slug=feedback_slug)
    if instance.of_tutor != request.user and not in_groups(request.user, ('Reviewers', )):
        raise Http404
    instance.unassign_tutor()
    instance.delete()
    return HttpResponseRedirect(reverse('start'))


def markfinal_feedback(request, feedback_slug):
    instance = Feedback.objects.get(slug=feedback_slug)
    instance.finalize_feedback(request.user)
    return HttpResponseRedirect(reverse('start'))


def markunfinal_feedback(request, feedback_slug):
    instance = Feedback.objects.get(slug=feedback_slug)
    instance.unfinalize_feedback()
    return HttpResponseRedirect(reverse('start'))


@group_required('Tutors', 'Reviewers')
def finished(request):
    return HttpResponse("Ich habe fertig.")


class FeedbackEdit(UpdateView):

    """docstring for FeedbackCreate"""

    form_class = FeedbackForm
    template_name = 'core/feedback_form.html'
    success_url = '/start/'

    @method_decorator(group_required('Tutors', 'Reviewers'))
    def dispatch(self, *args, **kwargs):
        return super(FeedbackEdit, self).dispatch(*args, **kwargs)

    def get_object(self):
        instance = Feedback.objects.get(slug=self.kwargs['feedback_slug'])
        if not instance.of_tutor == self.request.user:
            raise Http404
        return instance

    def form_valid(self, form):
        """
        If the form is valid, redirect to the supplied URL.
        """
        if form.is_valid() and not form.instance.final:
            form.instance.empty = False
            form.save()
        return HttpResponseRedirect(self.get_success_url())


class SubmissionView(DetailView):

    template_name = 'core/submission_view.html'
    model = Submission

    @method_decorator(group_required('Reviewers', 'Students'))
    def dispatch(self, *args, **kwargs):
        return super(SubmissionView, self).dispatch(*args, **kwargs)

    def get_object(self):
        return Submission.objects.get(slug=self.kwargs['slug'])
